/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
extern SLang_Intrin_Fun_Type JedLine_Intrinsics[];
extern void jed_skip_hidden_lines_backward (int *);
extern void jed_skip_hidden_lines_forward (int *);
