/* Copyright (c) 1992, 1998, 2000, 2005, 2006 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
#define PC_NULL	"\xE0^C"
#define PC_SLASH	"\033OQ"
#define PC_STAR		"\033OR"
#define PC_MINUS	"\033OS"
#define PC_PLUS		"\033Om"
#define PC_ENTER	"\033OM"
#define PC_KP5		"\033Ou"
#define PC_UP		"\xE0H"
#define PC_UP1		"\033Ox"
#define PC_DN		"\xE0P"
#define PC_DN1		"\033Or"
#define PC_RT		"\xE0M"
#define PC_RT1		"\033Ov"
#define PC_LT		"\xE0K"
#define PC_LT1		"\033Ot"
#define PC_PGUP		"\xE0I"
#define PC_PGUP1	"\033Oy"
#define PC_PGDN		"\xE0Q"
#define PC_PGDN1	"\033Os"
#define PC_INS		"\xE0R"
#define PC_INS1		"\033Op"
#define PC_DEL		"\xE0S"
#define PC_DEL1		"\033On"
#define PC_END		"\xE0O"
#define PC_END1		"\033Oq"
#define PC_HOME		"\xE0G"
#define PC_HOME1	"\033Ow"

#define PC_F1	"^@;"
#define PC_F2	"^@<"
#define PC_F3	"^@="
#define PC_F4	"^@>"
#define PC_F5	"^@?"
#define PC_F6	"^@@"
#define PC_F7	"^@A"
#define PC_F8	"^@B"
#define PC_F9	"^@C"
#define PC_F10	"^@D"

#define PC_ALT_F1	"^@h"
#define PC_ALT_F2	"^@i"
#define PC_ALT_F3	"^@j"
#define PC_ALT_F4	"^@k"
#define PC_ALT_F5	"^@l"
#define PC_ALT_F6	"^@m"
#define PC_ALT_F7	"^@n"
#define PC_ALT_F8	"^@o"
#define PC_ALT_F9	"^@p"
#define PC_ALT_F10	"^@q"

#define PC_SHIFT_F1	"^@T"
#define PC_SHIFT_F2	"^@U"
#define PC_SHIFT_F3	"^@V"
#define PC_SHIFT_F4	"^@W"
#define PC_SHIFT_F5	"^@X"
#define PC_SHIFT_F6	"^@Y"
#define PC_SHIFT_F7	"^@Z"
#define PC_SHIFT_F8	"^@["
#define PC_SHIFT_F9	"^@\\"
#define PC_SHIFT_F10	"^@]"
