#ifndef JED_VERSION
#define JED_VERSION	9919

#define JED_VERSION_STR "0.99.19"

/* Copyright (c) 1999, 2000, 2001, 2005-2009 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */

extern int Jed_Version_Number;
extern char *Jed_Version_String;

extern void jed_show_version (FILE *);

#endif				       /* JED_VERSION */
