/* Copyright (c) 1999, 2000, 2005, 2006 John E. Davis
 * This file is part of JED editor library source.
 *
 * You may distribute this file under the terms the GNU General Public
 * License.  See the file COPYING for more information.
 */
extern int jed_init_userinfo (void);
extern char *jed_get_hostname (void);
extern char *jed_get_username (void);
